#' Data about thyroid hormones for anonymous patients
#'
#' This data contains several obvious outliers from misspellings in data entry.
#' From Garavan Institute. For more details see link at the bottom.
#' 
#' @docType data
#' @usage data(hypothyroid)
#' @source \url{http://archive.ics.uci.edu/ml/datasets/thyroid+disease}
"hypothyroid"